"""Unit tests for servicecatalog-supported APIs."""

import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_portfolio():
    """Test create_portfolio API."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[{"Key": "testkey", "Value": "testvalue"}],
        IdempotencyToken="test-token",
    )

    assert "PortfolioDetail" in response
    assert "Tags" in response

    portfolio_detail = response["PortfolioDetail"]
    assert "Id" in portfolio_detail
    assert "ARN" in portfolio_detail
    assert "DisplayName" in portfolio_detail
    assert "Description" in portfolio_detail
    assert "CreatedTime" in portfolio_detail
    assert "ProviderName" in portfolio_detail

    assert portfolio_detail["DisplayName"] == "Test Portfolio"
    assert portfolio_detail["Description"] == "Test Portfolio Description"
    assert portfolio_detail["ProviderName"] == "Test Provider"

    tags = response["Tags"]
    assert len(tags) == 1
    assert tags[0]["Key"] == "testkey"
    assert tags[0]["Value"] == "testvalue"

    response2 = client.create_portfolio(
        DisplayName="Different Name",
        Description="Different Description",
        ProviderName="Different Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    assert response["PortfolioDetail"]["Id"] == response2["PortfolioDetail"]["Id"]
    assert (
        response["PortfolioDetail"]["DisplayName"]
        == response2["PortfolioDetail"]["DisplayName"]
    )


@mock_aws
def test_delete_portfolio():
    """Test delete_portfolio API."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    create_response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    portfolio_id = create_response["PortfolioDetail"]["Id"]

    list_response_before = client.list_portfolios()
    assert "PortfolioDetails" in list_response_before
    portfolio_ids_before = [p["Id"] for p in list_response_before["PortfolioDetails"]]
    assert portfolio_id in portfolio_ids_before

    client.delete_portfolio(Id=portfolio_id)

    list_response_after = client.list_portfolios()
    assert "PortfolioDetails" in list_response_after
    portfolio_ids_after = [p["Id"] for p in list_response_after["PortfolioDetails"]]
    assert portfolio_id not in portfolio_ids_after


@mock_aws
def test_create_portfolio_share_with_account():
    """Test create_portfolio_share API with an account ID."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    create_response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    portfolio_id = create_response["PortfolioDetail"]["Id"]

    share_response = client.create_portfolio_share(
        PortfolioId=portfolio_id, AccountId="123456789012"
    )

    assert "PortfolioShareToken" not in share_response

    access_response = client.list_portfolio_access(PortfolioId=portfolio_id)

    assert "AccountIds" in access_response
    assert "123456789012" in access_response["AccountIds"]


@mock_aws
def test_create_portfolio_share_with_organization():
    """Test create_portfolio_share API with an organization node."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    create_response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    portfolio_id = create_response["PortfolioDetail"]["Id"]

    share_response = client.create_portfolio_share(
        PortfolioId=portfolio_id,
        OrganizationNode={"Type": "ORGANIZATION", "Value": "o-exampleorgid"},
        ShareTagOptions=True,
        SharePrincipals=True,
    )

    assert share_response["PortfolioShareToken"].startswith(f"share-{portfolio_id}")

    assert "ORGANIZATION" in share_response["PortfolioShareToken"]
    assert "o-exampleorgid" in share_response["PortfolioShareToken"]
    assert "tags" in share_response["PortfolioShareToken"]
    assert "principals" in share_response["PortfolioShareToken"]


@mock_aws
def test_delete_portfolio_share_with_account():
    """Test delete_portfolio_share API with an account ID."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    create_response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    portfolio_id = create_response["PortfolioDetail"]["Id"]

    client.create_portfolio_share(PortfolioId=portfolio_id, AccountId="123456789012")

    access_response = client.list_portfolio_access(PortfolioId=portfolio_id)
    assert "123456789012" in access_response["AccountIds"]

    client.delete_portfolio_share(PortfolioId=portfolio_id, AccountId="123456789012")

    access_response = client.list_portfolio_access(PortfolioId=portfolio_id)
    assert "123456789012" not in access_response["AccountIds"]


@mock_aws
def test_delete_portfolio_share_with_organization():
    """Test delete_portfolio_share API with an organization node."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    create_response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    portfolio_id = create_response["PortfolioDetail"]["Id"]

    delete_response = client.delete_portfolio_share(
        PortfolioId=portfolio_id,
        OrganizationNode={"Type": "ORGANIZATION", "Value": "o-exampleorgid"},
    )

    assert delete_response["PortfolioShareToken"].startswith(f"share-{portfolio_id}")
    assert "ORGANIZATION" in delete_response["PortfolioShareToken"]
    assert "o-exampleorgid" in delete_response["PortfolioShareToken"]


@mock_aws
def test_list_portfolio_access():
    """Test list_portfolio_access API."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    create_response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    portfolio_id = create_response["PortfolioDetail"]["Id"]

    client.create_portfolio_share(PortfolioId=portfolio_id, AccountId="111111111111")

    client.create_portfolio_share(PortfolioId=portfolio_id, AccountId="222222222222")

    access_response = client.list_portfolio_access(PortfolioId=portfolio_id)

    assert "111111111111" in access_response["AccountIds"]
    assert "222222222222" in access_response["AccountIds"]

    paginated_response = client.list_portfolio_access(
        PortfolioId=portfolio_id, PageSize=1
    )

    assert len(paginated_response["AccountIds"]) == 1
    assert "NextPageToken" in paginated_response

    if paginated_response.get("NextPageToken"):
        next_page = client.list_portfolio_access(
            PortfolioId=portfolio_id,
            PageSize=1,
            PageToken=paginated_response["NextPageToken"],
        )

        assert len(next_page["AccountIds"]) == 1
        assert next_page["AccountIds"][0] != paginated_response["AccountIds"][0]


@mock_aws
def test_list_portfolios():
    """Test list_portfolios API."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    client.create_portfolio(
        DisplayName="Test Portfolio 1",
        Description="Test Portfolio 1 Description",
        ProviderName="Test Provider 1",
        Tags=[],
        IdempotencyToken="test-token-1",
    )

    client.create_portfolio(
        DisplayName="Test Portfolio 2",
        Description="Test Portfolio 2 Description",
        ProviderName="Test Provider 2",
        Tags=[],
        IdempotencyToken="test-token-2",
    )

    response = client.list_portfolios()

    assert "PortfolioDetails" in response

    portfolio_details = response["PortfolioDetails"]
    assert len(portfolio_details) == 2

    for portfolio in portfolio_details:
        assert "Id" in portfolio
        assert "ARN" in portfolio
        assert "DisplayName" in portfolio
        assert "Description" in portfolio
        assert "CreatedTime" in portfolio
        assert "ProviderName" in portfolio

    paginated_response = client.list_portfolios(PageSize=1)
    assert len(paginated_response["PortfolioDetails"]) == 2


@mock_aws
def test_describe_portfolio_shares():
    """Test describe_portfolio_shares API."""
    client = boto3.client("servicecatalog", region_name="us-east-1")

    create_response = client.create_portfolio(
        DisplayName="Test Portfolio",
        Description="Test Portfolio Description",
        ProviderName="Test Provider",
        Tags=[],
        IdempotencyToken="test-token",
    )

    portfolio_id = create_response["PortfolioDetail"]["Id"]

    client.create_portfolio_share(PortfolioId=portfolio_id, AccountId="111111111111")
    client.create_portfolio_share(PortfolioId=portfolio_id, AccountId="222222222222")
    client.create_portfolio_share(PortfolioId=portfolio_id, AccountId="333333333333")

    client.create_portfolio_share(
        PortfolioId=portfolio_id,
        OrganizationNode={"Type": "ORGANIZATION", "Value": "o-exampleorgid"},
        ShareTagOptions=True,
        SharePrincipals=True,
    )

    account_response = client.describe_portfolio_shares(
        PortfolioId=portfolio_id, Type="ACCOUNT"
    )

    assert "PortfolioShareDetails" in account_response
    assert len(account_response["PortfolioShareDetails"]) == 3

    for share in account_response["PortfolioShareDetails"]:
        assert share["Type"] == "ACCOUNT"
        assert share["Accepted"] is True
        assert share["PrincipalId"] in ["111111111111", "222222222222", "333333333333"]

    org_response = client.describe_portfolio_shares(
        PortfolioId=portfolio_id, Type="ORGANIZATION"
    )

    assert "PortfolioShareDetails" in org_response
    assert len(org_response["PortfolioShareDetails"]) == 1

    org_share = org_response["PortfolioShareDetails"][0]
    assert org_share["Type"] == "ORGANIZATION"
    assert org_share["PrincipalId"] == "o-exampleorgid"
    assert org_share["Accepted"] is True
    assert org_share["ShareTagOptions"] is True
    assert org_share["SharePrincipals"] is True

    client.delete_portfolio_share(
        PortfolioId=portfolio_id,
        OrganizationNode={"Type": "ORGANIZATION", "Value": "o-exampleorgid"},
    )

    org_response_after_delete = client.describe_portfolio_shares(
        PortfolioId=portfolio_id, Type="ORGANIZATION"
    )

    assert len(org_response_after_delete["PortfolioShareDetails"]) == 0



import boto3

from moto import mock_aws


@mock_aws
def test_create_graphql_api_with_tags():
    client = boto3.client("appsync", region_name="ap-southeast-1")
    api = client.create_graphql_api(
        name="api1", authenticationType="API_KEY", tags={"key": "val", "key2": "val2"}
    )["graphqlApi"]

    assert api["tags"] == {"key": "val", "key2": "val2"}

    api = client.get_graphql_api(apiId=api["apiId"])["graphqlApi"]

    assert api["tags"] == {"key": "val", "key2": "val2"}


@mock_aws
def test_tag_resource():
    client = boto3.client("appsync", region_name="us-east-2")
    api = client.create_graphql_api(name="api1", authenticationType="API_KEY")[
        "graphqlApi"
    ]

    client.tag_resource(resourceArn=(api["arn"]), tags={"key1": "val1"})

    api = client.get_graphql_api(apiId=api["apiId"])["graphqlApi"]
    assert api["tags"] == {"key1": "val1"}


@mock_aws
def test_tag_resource_with_existing_tags():
    client = boto3.client("appsync", region_name="us-east-2")
    api = client.create_graphql_api(
        name="api1", authenticationType="API_KEY", tags={"key": "val", "key2": "val2"}
    )["graphqlApi"]

    client.untag_resource(resourceArn=api["arn"], tagKeys=["key"])

    client.tag_resource(
        resourceArn=(api["arn"]), tags={"key2": "new value", "key3": "val3"}
    )

    api = client.get_graphql_api(apiId=api["apiId"])["graphqlApi"]
    assert api["tags"] == {"key2": "new value", "key3": "val3"}


@mock_aws
def test_untag_resource():
    client = boto3.client("appsync", region_name="eu-west-1")
    api = client.create_graphql_api(
        name="api1", authenticationType="API_KEY", tags={"key": "val", "key2": "val2"}
    )["graphqlApi"]

    client.untag_resource(resourceArn=api["arn"], tagKeys=["key"])

    api = client.get_graphql_api(apiId=api["apiId"])["graphqlApi"]
    assert api["tags"] == {"key2": "val2"}


@mock_aws
def test_untag_resource_all():
    client = boto3.client("appsync", region_name="eu-west-1")
    api = client.create_graphql_api(
        name="api1", authenticationType="API_KEY", tags={"key": "val", "key2": "val2"}
    )["graphqlApi"]

    client.untag_resource(resourceArn=api["arn"], tagKeys=["key", "key2"])

    api = client.get_graphql_api(apiId=api["apiId"])["graphqlApi"]
    assert api["tags"] == {}


@mock_aws
def test_list_tags_for_resource():
    client = boto3.client("appsync", region_name="ap-southeast-1")
    api = client.create_graphql_api(
        name="api1", authenticationType="API_KEY", tags={"key": "val", "key2": "val2"}
    )["graphqlApi"]

    resp = client.list_tags_for_resource(resourceArn=api["arn"])

    assert resp["tags"] == {"key": "val", "key2": "val2"}


@mock_aws
def test_resource_groups_tagging_api():
    region = "us-west-2"
    client = boto3.client("appsync", region_name=region)
    rtapi_client = boto3.client("resourcegroupstaggingapi", region_name=region)

    api = client.create_graphql_api(
        name="api-with-tags",
        authenticationType="API_KEY",
        tags={"key": "val", "key2": "val2"},
    )["graphqlApi"]
    client.create_graphql_api(name="api-no-tags", authenticationType="API_KEY")[
        "graphqlApi"
    ]

    resp = rtapi_client.get_resources(ResourceTypeFilters=["appsync"])
    assert len(resp["ResourceTagMappingList"]) == 1
    assert resp["ResourceTagMappingList"][0]["ResourceARN"] == api["arn"]
    assert resp["ResourceTagMappingList"][0]["Tags"] == [
        {"Key": "key", "Value": "val"},
        {"Key": "key2", "Value": "val2"},
    ]



import uuid

from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from moto.wafv2.utils import make_arn_for_wacl


def test_make_arn_for_wacl_in_regional_scope():
    uniqueID = str(uuid.uuid4())
    region = "us-east-1"
    name = "testName"
    scope = "REGIONAL"

    arn = make_arn_for_wacl(name, ACCOUNT_ID, region, uniqueID, scope)

    assert (
        arn == f"arn:aws:wafv2:{region}:{ACCOUNT_ID}:regional/webacl/{name}/{uniqueID}"
    )


def test_make_arn_for_wacl_in_cloudfront_scope():
    uniqueID = str(uuid.uuid4())
    backend_region = "aws"  # see https://github.com/getmoto/moto/blob/d00aa025b6c3d37977508b5d5e81ecad4ca15159/moto/core/base_backend.py#L272
    arn_region = "us-east-1"
    name = "testName"
    scope = "CLOUDFRONT"

    arn = make_arn_for_wacl(name, ACCOUNT_ID, backend_region, uniqueID, scope)

    assert (
        arn
        == f"arn:aws:wafv2:{arn_region}:{ACCOUNT_ID}:global/webacl/{name}/{uniqueID}"
    )



from __future__ import annotations

from collections import defaultdict
from functools import lru_cache
from typing import Any, Dict, List, Optional

from botocore import xform_name

from moto.core.utils import get_service_model


class XFormedAttributeAccessMixin:
    """Mixin allowing access to "xformed" attributes:

    obj.DBInstanceIdentifier will retrieve the value of obj.db_instance_identifier

    """

    BOTOCORE_MODEL: Optional[str] = None

    _model_attribute_aliases: Dict[str, List[str]] = {}
    _xform_cache: Dict[str, str] = {}

    def __getattr__(self, name: str) -> Any:
        if name in self.model_attributes:
            return self.get_modeled_attribute(name)
        raise AttributeError(f"Attribute '{name}' not found!")

    def get_modeled_attribute(self, attr_name: str) -> Any:
        for attr_alias in self.model_attribute_aliases[attr_name]:
            try:
                return super().__getattribute__(attr_alias)
            except AttributeError:
                pass
        else:
            raise AttributeError

    @property
    def model_attributes(self) -> List[str]:
        return list(self.model_attribute_aliases.keys())

    @property
    def model_attribute_aliases(self) -> Dict[str, List[str]]:
        if not self._model_attribute_aliases:
            self._model_attribute_aliases = self.get_model_attributes_info()
        return self._model_attribute_aliases

    @classmethod
    @lru_cache()
    def get_model_attributes_info(cls) -> Dict[str, List[str]]:
        service_name = cls.__module__.split(".")[1]
        model_name = cls.BOTOCORE_MODEL or cls.__name__
        service_model = get_service_model(service_name)
        model_shape = service_model.shape_for(model_name)
        valid_attributes: Dict[str, List[str]] = defaultdict(list)
        for member_name, member_shape in model_shape.members.items():  # type: ignore[attr-defined]
            aliases = valid_attributes[member_name]
            if member_shape.type_name == "list":
                if member_name != member_shape.name:
                    xformed_name = cls._xform_name(member_shape.name)
                    aliases.append(xformed_name)
            xformed_member_name = cls._xform_name(member_name)
            aliases.append(xformed_member_name)
            if member_name.startswith(model_name):
                short_name = member_name[len(model_name) :]
                xformed_short_name = cls._xform_name(short_name)
                aliases.append(xformed_short_name)
        return valid_attributes

    @classmethod
    def _xform_name(cls, name: str) -> str:
        return xform_name(name, _xform_cache=cls._xform_cache)



"""Unit tests for kinesisanalyticsv2-supported APIs."""

import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html

FAKE_SECURITY_GROUP_IDS = ["sg-0123456789abcdef0"]
FAKE_SUBNET_IDS = ["subnet-0123456789abcdef0", "subnet-abcdef0123456789"]
FAKE_TAGS = [
    {"Key": "TestKey", "Value": "TestValue"},
    {"Key": "TestKey2", "Value": "TestValue2"},
]
FAKE_VPC_ID = "vpc-0123456789abcdef0"
FAKE_BUCKET_ARN = "arn:aws:s3:::test"
FAKE_FILE_KEY = "testfile.jar"


@mock_aws
def test_create_application():
    region = "us-east-2"
    client = boto3.client("kinesisanalyticsv2", region_name=region)
    log_stream = f"arn:aws:logs:{region}:{ACCOUNT_ID}:log-group:/aws/kinesis-analytics/log-stream:test-log-stream"

    resp = client.create_application(
        ApplicationName="test_application",
        ApplicationDescription="test application description",
        RuntimeEnvironment="FLINK-1_20",
        ServiceExecutionRole=f"arn:aws:iam::{ACCOUNT_ID}:role/application_role",
        Tags=FAKE_TAGS,
        CloudWatchLoggingOptions=[{"LogStreamARN": log_stream}],
    )

    app = resp.get("ApplicationDetail")
    app_arn = app.get("ApplicationARN")
    assert app_arn
    assert app.get("ApplicationDescription") == "test application description"

    assert app.get("RuntimeEnvironment") == "FLINK-1_20"
    assert (
        app.get("ServiceExecutionRole")
        == f"arn:aws:iam::{ACCOUNT_ID}:role/application_role"
    )

    tags_resp = client.list_tags_for_resource(ResourceARN=app_arn)
    assert len(tags_resp.get("Tags")) == 2
    assert tags_resp.get("Tags") == FAKE_TAGS
    assert len(app.get("CloudWatchLoggingOptionDescriptions")[0]) == 2


@mock_aws
def test_create_application_with_appconfig():
    region = "us-east-2"
    client = boto3.client("kinesisanalyticsv2", region_name=region)

    resp = client.create_application(
        ApplicationName="test_application",
        RuntimeEnvironment="FLINK-1_20",
        ServiceExecutionRole=f"arn:aws:iam::{ACCOUNT_ID}:role/application_role",
        ApplicationConfiguration={
            "FlinkApplicationConfiguration": {
                "CheckpointConfiguration": {"ConfigurationType": "DEFAULT"},
                "MonitoringConfiguration": {"ConfigurationType": "DEFAULT"},
                "ParallelismConfiguration": {"ConfigurationType": "DEFAULT"},
            },
            "EnvironmentProperties": {
                "PropertyGroups": [
                    {
                        "PropertyGroupId": "TEST",
                        "PropertyMap": {
                            "aws.region": "us-east-2",
                            "AggregationEnabled": "false",
                        },
                    },
                    {
                        "PropertyGroupId": "TEST2",
                        "PropertyMap": {
                            "aws.region": "us-west-2",
                        },
                    },
                ]
            },
            "ApplicationCodeConfiguration": {
                "CodeContent": {
                    "S3ContentLocation": {
                        "BucketARN": FAKE_BUCKET_ARN,
                        "FileKey": FAKE_FILE_KEY,
                        "ObjectVersion": "1",
                    }
                },
                "CodeContentType": "ZIPFILE",
            },
            "ApplicationSnapshotConfiguration": {"SnapshotsEnabled": False},
            "ApplicationSystemRollbackConfiguration": {"RollbackEnabled": False},
            "VpcConfigurations": [
                {"SubnetIds": FAKE_SUBNET_IDS, "SecurityGroupIds": FAKE_SUBNET_IDS}
            ],
            "ZeppelinApplicationConfiguration": {
                "MonitoringConfiguration": {"LogLevel": "INFO"},
                "CatalogConfiguration": {
                    "GlueDataCatalogConfiguration": {
                        "DatabaseARN": f"arn:aws:glue:{region}:{ACCOUNT_ID}:database/test"
                    }
                },
                "DeployAsApplicationConfiguration": {
                    "S3ContentLocation": {
                        "BucketARN": FAKE_BUCKET_ARN,
                        "BasePath": "test/app",
                    }
                },
                "CustomArtifactsConfiguration": [
                    {
                        "ArtifactType": "DEPENDENCY_JAR",
                        "S3ContentLocation": {
                            "BucketARN": FAKE_BUCKET_ARN,
                            "FileKey": FAKE_FILE_KEY,
                            "ObjectVersion": "1.0",
                        },
                        "MavenReference": {
                            "GroupId": "org.apache.flink",
                            "ArtifactId": "flink-connector-kafka_2.12",
                            "Version": "1.13.2",
                        },
                    },
                ],
            },
        },
    )
    app = resp.get("ApplicationDetail")
    app_config = app.get("ApplicationConfigurationDescription")

    assert app_config.get("FlinkApplicationConfigurationDescription") == {
        "CheckpointConfigurationDescription": {
            "ConfigurationType": "DEFAULT",
            "CheckpointingEnabled": True,
            "CheckpointInterval": 60000,
            "MinPauseBetweenCheckpoints": 5000,
        },
        "MonitoringConfigurationDescription": {
            "ConfigurationType": "DEFAULT",
            "MetricsLevel": "APPLICATION",
            "LogLevel": "INFO",
        },
        "ParallelismConfigurationDescription": {
            "ConfigurationType": "DEFAULT",
            "Parallelism": 1,
            "ParallelismPerKPU": 1,
            "AutoScalingEnabled": False,
            "CurrentParallelism": 1,
        },
    }

    assert app_config.get("EnvironmentPropertyDescriptions") == {
        "PropertyGroupDescriptions": [
            {
                "PropertyGroupId": "TEST",
                "PropertyMap": {
                    "aws.region": "us-east-2",
                    "AggregationEnabled": "false",
                },
            },
            {
                "PropertyGroupId": "TEST2",
                "PropertyMap": {
                    "aws.region": "us-west-2",
                },
            },
        ]
    }

    assert app_config == {
        "FlinkApplicationConfigurationDescription": {
            "CheckpointConfigurationDescription": {
                "ConfigurationType": "DEFAULT",
                "CheckpointingEnabled": True,
                "CheckpointInterval": 60000,
                "MinPauseBetweenCheckpoints": 5000,
            },
            "MonitoringConfigurationDescription": {
                "ConfigurationType": "DEFAULT",
                "MetricsLevel": "APPLICATION",
                "LogLevel": "INFO",
            },
            "ParallelismConfigurationDescription": {
                "ConfigurationType": "DEFAULT",
                "Parallelism": 1,
                "ParallelismPerKPU": 1,
                "AutoScalingEnabled": False,
                "CurrentParallelism": 1,
            },
        },
        "EnvironmentPropertyDescriptions": {
            "PropertyGroupDescriptions": [
                {
                    "PropertyGroupId": "TEST",
                    "PropertyMap": {
                        "aws.region": "us-east-2",
                        "AggregationEnabled": "false",
                    },
                },
                {
                    "PropertyGroupId": "TEST2",
                    "PropertyMap": {
                        "aws.region": "us-west-2",
                    },
                },
            ]
        },
        "ApplicationCodeConfigurationDescription": {
            "CodeContentDescription": {
                "CodeMD5": "fakechecksum",
                "CodeSize": 123,
                "S3ApplicationCodeLocationDescription": {
                    "BucketARN": FAKE_BUCKET_ARN,
                    "FileKey": FAKE_FILE_KEY,
                    "ObjectVersion": "1",
                },
            },
            "CodeContentType": "ZIPFILE",
        },
        "ApplicationSnapshotConfigurationDescription": {"SnapshotsEnabled": False},
        "ApplicationSystemRollbackConfigurationDescription": {"RollbackEnabled": False},
        "VpcConfigurationDescriptions": [
            {
                "VpcConfigurationId": "1.1",
                "VpcId": FAKE_VPC_ID,
                "SubnetIds": FAKE_SUBNET_IDS,
                "SecurityGroupIds": FAKE_SUBNET_IDS,
            }
        ],
        "ZeppelinApplicationConfigurationDescription": {
            "MonitoringConfigurationDescription": {"LogLevel": "INFO"},
            "CatalogConfigurationDescription": {
                "GlueDataCatalogConfigurationDescription": {
                    "DatabaseARN": f"arn:aws:glue:{region}:{ACCOUNT_ID}:database/test"
                }
            },
            "DeployAsApplicationConfigurationDescription": {
                "S3ContentLocationDescription": {
                    "BucketARN": FAKE_BUCKET_ARN,
                    "BasePath": "test/app",
                }
            },
            "CustomArtifactsConfigurationDescription": [
                {
                    "ArtifactType": "DEPENDENCY_JAR",
                    "S3ContentLocationDescription": {
                        "BucketARN": FAKE_BUCKET_ARN,
                        "FileKey": FAKE_FILE_KEY,
                        "ObjectVersion": "1.0",
                    },
                    "MavenReferenceDescription": {
                        "GroupId": "org.apache.flink",
                        "ArtifactId": "flink-connector-kafka_2.12",
                        "Version": "1.13.2",
                    },
                },
            ],
        },
    }

    # Test CUSTOM flink app configurations
    flink_custom_resp = client.create_application(
        ApplicationName="test_application",
        RuntimeEnvironment="FLINK-1_20",
        ServiceExecutionRole=f"arn:aws:iam::{ACCOUNT_ID}:role/application_role",
        ApplicationConfiguration={
            "FlinkApplicationConfiguration": {
                "CheckpointConfiguration": {
                    "ConfigurationType": "CUSTOM",
                    "CheckpointingEnabled": True,
                    "CheckpointInterval": 12345,
                    "MinPauseBetweenCheckpoints": 1234,
                },
                "MonitoringConfiguration": {
                    "ConfigurationType": "CUSTOM",
                    "MetricsLevel": "OPERATOR",
                    "LogLevel": "ERROR",
                },
                "ParallelismConfiguration": {
                    "ConfigurationType": "CUSTOM",
                    "AutoScalingEnabled": True,
                },
            }
        },
    )

    flink_app = flink_custom_resp.get("ApplicationDetail")
    flink_app_config = flink_app.get("ApplicationConfigurationDescription")
    # Custom should use default values if values are not provided
    assert flink_app_config.get("FlinkApplicationConfigurationDescription") == {
        "CheckpointConfigurationDescription": {
            "ConfigurationType": "CUSTOM",
            "CheckpointingEnabled": True,
            "CheckpointInterval": 12345,
            "MinPauseBetweenCheckpoints": 1234,
        },
        "MonitoringConfigurationDescription": {
            "ConfigurationType": "CUSTOM",
            "MetricsLevel": "OPERATOR",
            "LogLevel": "ERROR",
        },
        "ParallelismConfigurationDescription": {
            "ConfigurationType": "CUSTOM",
            "Parallelism": 1,
            "ParallelismPerKPU": 1,
            "AutoScalingEnabled": True,
            "CurrentParallelism": 1,
        },
    }


@mock_aws
def test_tag_resource():
    region = "us-east-2"
    client = boto3.client("kinesisanalyticsv2", region_name=region)
    app_resp = client.create_application(
        ApplicationName="test_application",
        RuntimeEnvironment="FLINK-1_20",
        ServiceExecutionRole=f"arn:aws:iam::{ACCOUNT_ID}:role/application_role",
    )
    app = app_resp.get("ApplicationDetail")
    app_arn = app.get("ApplicationARN")
    client.tag_resource(
        ResourceARN=app_arn,
        Tags=[
            {"Key": "key2", "Value": "value2"},
        ],
    )

    tags_resp = client.list_tags_for_resource(ResourceARN=app_arn)
    assert tags_resp.get("Tags") == [{"Key": "key2", "Value": "value2"}]


@mock_aws
def test_describe_application():
    region = "us-east-2"
    client = boto3.client("kinesisanalyticsv2", region_name=region)
    client.create_application(
        ApplicationName="test_application",
        RuntimeEnvironment="FLINK-1_20",
        ServiceExecutionRole=f"arn:aws:iam::{ACCOUNT_ID}:role/application_role",
    )
    app_resp = client.describe_application(ApplicationName="test_application")

    app = app_resp.get("ApplicationDetail")
    assert (
        app.get("ApplicationARN")
        == f"arn:aws:kinesisanalytics:{region}:{ACCOUNT_ID}:application/test_application"
    )
    assert app.get("RuntimeEnvironment") == "FLINK-1_20"
    assert (
        app.get("ServiceExecutionRole")
        == f"arn:aws:iam::{ACCOUNT_ID}:role/application_role"
    )


@mock_aws
def test_list_applications():
    region = "us-east-2"
    client = boto3.client("kinesisanalyticsv2", region_name=region)

    client.create_application(
        ApplicationName="test_application",
        RuntimeEnvironment="FLINK-1_20",
        ServiceExecutionRole=f"arn:aws:iam::{ACCOUNT_ID}:role/application_role",
    )

    resp = client.list_applications()
    app_summaries = resp.get("ApplicationSummaries")

    assert len(app_summaries) == 1
    assert app_summaries[0].get("ApplicationName") == "test_application"
    assert (
        app_summaries[0].get("ApplicationARN")
        == f"arn:aws:kinesisanalytics:{region}:{ACCOUNT_ID}:application/test_application"
    )
    assert app_summaries[0].get("ApplicationStatus") == "STARTING"
    assert app_summaries[0].get("ApplicationVersionId") == 1
    assert app_summaries[0].get("RuntimeEnvironment") == "FLINK-1_20"



"""servicecatalog base URL and path."""

from .responses import ServiceCatalogResponse

url_bases = [
    r"https?://servicecatalog\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": ServiceCatalogResponse.dispatch,
}



# mypy: ignore-errors
"""Additional tests for response serialization.

While there are compliance tests in ``test_protocols.py``, where the majority
of the response serialization is tested, this file contains additional tests
that go above and beyond the specification(s) for a number of reasons:

* We are testing Python-specific behavior that doesn't make sense as a
  compliance test.
* We are testing behavior that is not strictly part of the specification.
  These may result in a coverage gap that would otherwise be untested.

"""

import time

from botocore.model import ServiceModel, ShapeResolver
from xmltodict import parse

from moto.core.serialize import QuerySerializer, XFormedAttributePicker


def test_serialize_from_object() -> None:
    model = {
        "metadata": {"protocol": "query", "apiVersion": "2014-01-01"},
        "documentation": "",
        "operations": {
            "TestOperation": {
                "name": "TestOperation",
                "http": {
                    "method": "POST",
                    "requestUri": "/",
                },
                "output": {"shape": "OutputShape"},
            }
        },
        "shapes": {
            "OutputShape": {
                "type": "structure",
                "members": {
                    "string": {"shape": "StringType"},
                },
            },
            "StringType": {
                "type": "string",
            },
        },
    }

    class TestObject:
        string = "test-string"

    service_model = ServiceModel(model)
    operation_model = service_model.operation_model("TestOperation")
    serializer = QuerySerializer(operation_model)
    serialized = serializer.serialize(TestObject())
    assert serialized


def test_datetime_with_microseconds() -> None:
    model = {
        "metadata": {"protocol": "query", "apiVersion": "2014-01-01"},
        "documentation": "",
        "operations": {
            "TestOperation": {
                "name": "TestOperation",
                "http": {
                    "method": "POST",
                    "requestUri": "/",
                },
                "output": {"shape": "OutputShape"},
            }
        },
        "shapes": {
            "OutputShape": {
                "type": "structure",
                "members": {
                    "microseconds": {"shape": "TimestampType"},
                },
            },
            "TimestampType": {
                "type": "timestamp",
            },
        },
    }

    class TestObject:
        microseconds = time.time()

    service_model = ServiceModel(model)
    operation_model = service_model.operation_model("TestOperation")
    serializer = QuerySerializer(operation_model)
    serialized = serializer.serialize(TestObject())
    assert serialized
    parsed = parse(serialized["body"])
    time_str = parsed["TestOperationResponse"]["OutputShapeResult"]["microseconds"]
    assert "." in time_str
    assert time_str[-1] == "Z"


def test_pretty_print_with_short_elements_and_list() -> None:
    model = {
        "metadata": {"protocol": "query", "apiVersion": "2014-01-01"},
        "documentation": "",
        "operations": {
            "TestOperation": {
                "name": "TestOperation",
                "http": {
                    "method": "POST",
                    "requestUri": "/",
                },
                "output": {"shape": "OutputShape"},
            }
        },
        "shapes": {
            "OutputShape": {
                "type": "structure",
                "members": {
                    "DBInstances": {
                        "shape": "DBInstanceList",
                    }
                },
            },
            "DBInstanceList": {
                "type": "list",
                "member": {"shape": "DBInstance", "locationName": "DBInstance"},
            },
            "DBInstance": {
                "type": "structure",
                "members": {
                    "DBInstanceIdentifier": {
                        "shape": "String",
                    }
                },
            },
            "String": {"type": "string"},
        },
    }
    service_model = ServiceModel(model)
    operation_model = service_model.operation_model("TestOperation")
    serializer = QuerySerializer(operation_model, **{"pretty_print": True})
    empty_list_to_serialize = {"DBInstances": []}
    serialized = serializer.serialize(empty_list_to_serialize)
    expected_shortened_list_element = "<DBInstances/>"
    assert expected_shortened_list_element in serialized["body"]


class TestXFormedAttributePicker:
    picker = XFormedAttributePicker()

    def test_missing_attribute(self):
        obj = dict()
        value = self.picker(obj, "Attribute", None)
        assert value is None

    def test_serialization_key(self):
        shape_map = {
            "Foo": {
                "type": "structure",
                "members": {
                    "Baz": {"shape": "StringType", "locationName": "other"},
                },
            },
            "StringType": {"type": "string"},
        }
        resolver = ShapeResolver(shape_map)
        shape = resolver.resolve_shape_ref(
            {"shape": "StringType", "locationName": "other"}
        )
        assert shape
        obj = {"other": "found me"}
        value = self.picker(obj, "StringType", shape)
        assert value == "found me"

    def test_short_key(self):
        class Role:
            id = "unique-identifier"

        value = self.picker(Role(), "RoleId", None)
        assert value == "unique-identifier"

    def test_transformed_key(self):
        class DBInstance:
            db_instance_class = "t2.medium"

        value = self.picker(DBInstance(), "DBInstanceClass", None)
        assert value == "t2.medium"

    def test_untransformed_key(self):
        obj = {"PascalCasedAttr": True}
        value = self.picker(obj, "PascalCasedAttr", None)
        assert value is True



import json
import time
from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import merge_dicts
from moto.iot.models import IoTBackend, iot_backends

from .exceptions import (
    ConflictException,
    InvalidRequestException,
    ResourceNotFoundException,
)


class FakeShadow(BaseModel):
    """See the specification:
    http://docs.aws.amazon.com/iot/latest/developerguide/thing-shadow-document-syntax.html
    """

    def __init__(
        self,
        desired: Optional[str],
        reported: Optional[str],
        requested_payload: Optional[Dict[str, Any]],
        version: int,
        deleted: bool = False,
    ):
        self.desired = desired
        self.reported = reported
        self.requested_payload = requested_payload
        self.version = version
        self.timestamp = int(time.time())
        self.deleted = deleted

        self.metadata_desired = self._create_metadata_from_state(
            self.desired, self.timestamp
        )
        self.metadata_reported = self._create_metadata_from_state(
            self.reported, self.timestamp
        )

    @classmethod
    def create_from_previous_version(  # type: ignore[misc]
        cls, previous_shadow: Optional["FakeShadow"], payload: Optional[Dict[str, Any]]
    ) -> "FakeShadow":
        """
        set None to payload when you want to delete shadow
        """
        version, previous_payload = (
            (previous_shadow.version + 1, previous_shadow.to_dict(include_delta=False))
            if previous_shadow
            else (1, {})
        )

        if payload is None:
            # if given payload is None, delete existing payload
            # this means the request was delete_thing_shadow
            shadow = FakeShadow(None, None, None, version, deleted=True)
            return shadow

        # Updates affect only the fields specified in the request state document.
        # Any field with a value of None is removed from the device's shadow.
        state_document = previous_payload.copy()
        merge_dicts(state_document, payload, remove_nulls=True)
        desired = state_document.get("state", {}).get("desired")
        reported = state_document.get("state", {}).get("reported")
        return FakeShadow(desired, reported, payload, version)

    @classmethod
    def parse_payload(cls, desired: Any, reported: Any) -> Any:  # type: ignore[misc]
        if not desired and not reported:
            delta = None
        elif reported is None and desired:
            delta = desired
        elif desired and reported:
            delta = cls._compute_delta_dict(desired, reported)
        else:
            delta = None
        return delta

    @classmethod
    def _compute_delta_dict(cls, desired: Any, reported: Any) -> Dict[str, Any]:  # type: ignore[misc]
        delta = {}
        for key, value in desired.items():
            delta_value = cls._compute_delta(reported.get(key), value)

            if delta_value is not None:
                delta[key] = delta_value
        return delta

    @classmethod
    def _compute_delta(cls, reported_value: Any, desired_value: Any) -> Any:  # type: ignore[misc]
        if reported_value == desired_value:
            return None

        if isinstance(desired_value, dict) and isinstance(reported_value, dict):
            return cls._compute_delta_dict(desired_value, reported_value)

        # Types are different, or
        # Both types are intrinsic values (str, int, etc), or
        # Both types are lists:
        #
        # Just return the desired value
        return desired_value

    def _create_metadata_from_state(self, state: Any, ts: Any) -> Any:
        """
        state must be desired or reported stype dict object
        replaces primitive type with {"timestamp": ts} in dict
        """
        if state is None:
            return None

        def _f(elem: Any, ts: Any) -> Any:
            if isinstance(elem, dict):
                return {_: _f(elem[_], ts) for _ in elem.keys()}
            if isinstance(elem, list):
                return [_f(_, ts) for _ in elem]
            return {"timestamp": ts}

        return _f(state, ts)

    def to_response_dict(self) -> Dict[str, Any]:
        desired = self.requested_payload["state"].get("desired", None)  # type: ignore
        reported = self.requested_payload["state"].get("reported", None)  # type: ignore

        payload = {}
        if desired is not None:
            payload["desired"] = desired
        if reported is not None:
            payload["reported"] = reported

        metadata = {}
        if desired is not None:
            metadata["desired"] = self._create_metadata_from_state(
                desired, self.timestamp
            )
        if reported is not None:
            metadata["reported"] = self._create_metadata_from_state(
                reported, self.timestamp
            )
        return {
            "state": payload,
            "metadata": metadata,
            "timestamp": self.timestamp,
            "version": self.version,
        }

    def to_dict(self, include_delta: bool = True) -> Dict[str, Any]:
        """returning nothing except for just top-level keys for now."""
        if self.deleted:
            return {"timestamp": self.timestamp, "version": self.version}
        delta = self.parse_payload(self.desired, self.reported)
        payload = {}
        if self.desired:
            payload["desired"] = self.desired
        if self.reported:
            payload["reported"] = self.reported
        if include_delta and delta:
            payload["delta"] = delta

        metadata = {}
        if self.metadata_desired is not None:
            metadata["desired"] = self.metadata_desired
        if self.metadata_reported is not None:
            metadata["reported"] = self.metadata_reported

        return {
            "state": payload,
            "metadata": metadata,
            "timestamp": self.timestamp,
            "version": self.version,
        }


class IoTDataPlaneBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.published_payloads: List[Tuple[str, bytes]] = list()

    @property
    def iot_backend(self) -> IoTBackend:
        return iot_backends[self.account_id][self.region_name]

    def update_thing_shadow(
        self, thing_name: str, payload: str, shadow_name: Optional[str]
    ) -> FakeShadow:
        """
        spec of payload:
          - need node `state`
          - state node must be an Object
          - State contains an invalid node: 'foo'
        """
        thing = self.iot_backend.describe_thing(thing_name)

        # validate
        try:
            _payload = json.loads(payload)
        except ValueError:
            raise InvalidRequestException("invalid json")
        if "state" not in _payload:
            raise InvalidRequestException("need node `state`")
        if not isinstance(_payload["state"], dict):
            raise InvalidRequestException("state node must be an Object")
        if any(_ for _ in _payload["state"].keys() if _ not in ["desired", "reported"]):
            raise InvalidRequestException("State contains an invalid node")

        thing_shadow = thing.thing_shadows.get(shadow_name)
        if "version" in _payload and thing_shadow.version != _payload["version"]:  # type: ignore
            raise ConflictException("Version conflict")
        new_shadow = FakeShadow.create_from_previous_version(thing_shadow, _payload)
        thing.thing_shadows[shadow_name] = new_shadow
        return new_shadow

    def get_thing_shadow(
        self, thing_name: str, shadow_name: Optional[str]
    ) -> FakeShadow:
        thing = self.iot_backend.describe_thing(thing_name)
        thing_shadow = thing.thing_shadows.get(shadow_name)

        if thing_shadow is None or thing_shadow.deleted:
            raise ResourceNotFoundException()
        return thing_shadow

    def delete_thing_shadow(
        self, thing_name: str, shadow_name: Optional[str]
    ) -> FakeShadow:
        thing = self.iot_backend.describe_thing(thing_name)
        thing_shadow = thing.thing_shadows.get(shadow_name)
        if thing_shadow is None:
            raise ResourceNotFoundException()
        payload = None
        new_shadow = FakeShadow.create_from_previous_version(thing_shadow, payload)
        thing.thing_shadows[shadow_name] = new_shadow
        return new_shadow

    def publish(self, topic: str, payload: bytes) -> None:
        self.published_payloads.append((topic, payload))

    def list_named_shadows_for_thing(self, thing_name: str) -> List[str]:
        thing = self.iot_backend.describe_thing(thing_name)
        return [name for name in thing.thing_shadows.keys() if name is not None]


iotdata_backends = BackendDict(IoTDataPlaneBackend, "iot-data")



from typing import List


def find_duplicates(input_list: List[str]) -> List[str]:
    input_list_copy = input_list.copy()
    while len(input_list_copy) > 0:
        # start at front of list
        item = input_list_copy.pop(0)
        if item in input_list_copy:
            return [item, item]
    return []


def find_path_overlaps(input_list: List[str]) -> List[str]:
    input_list_copy = input_list.copy()
    while len(input_list_copy) > 0:
        # start at front of list
        item = input_list_copy.pop(0)
        if item in input_list_copy:
            return [item, item]
        else:
            for second_item in input_list_copy:
                if second_item.startswith(item + "."):
                    return [item, second_item]
                elif item.startswith(second_item + "."):
                    return [item, second_item]
    return []



# mypy: disable-error-code="misc, override, var-annotated"
"""Response serializers for the various AWS protocol specifications.

There are some similarities among the different protocols with respect
to response serialization, so the code is structured in a way to avoid
code duplication where possible.  The diagram below illustrates the
inheritance hierarchy of the response serializer classes.

                           +--------------------+
                           | ResponseSerializer |
                           +--------------------+
                             ^       ^       ^
         +-------------------+       |       +---------------------+
         |                           |                             |
    +----+--------------+  +---------+----------+  +---------------+----+
    | BaseXMLSerializer |  | BaseRestSerializer |  | BaseJSONSerializer |
    +-------------------+  +--------------------+  +--------------------+
         ^             ^          ^           ^           ^           ^
         |             |          |           |           |           |
         |         +---+----------+----+  +---+-----------+----+      |
         |         | RestXMLSerializer |  | RestJSONSerializer |      |
         |         +-------------------+  +--------------------+      |
         |                                                            |
    +----+------------+                                  +------------+---+
    | QuerySerializer |                                  | JSONSerializer |
    +-----------------+                                  +----------------+
         ^
         |
    +----+----------+
    | EC2Serializer |
    +---------------+

Return Value
============

The response serializers expose a single public method: ``serialize()``
This method takes in any result (dict, object, Exception, etc.) and
returns a serialized ResponseDict of the following form:

    {
        "body": <RESPONSE_BODY>,
        "headers": <RESPONSE_HEADERS>,
        "status_code": <RESPONSE_HTTP_STATUS_CODE>,
    }

The body serialization output is text (Python strings), not bytes, mostly
for ease of inspection while debugging (pretty printing is also supported).
The exception is blob types, which are assumed to be binary and will be
encoded as ``utf-8``.

"""

from __future__ import annotations

import base64
import calendar
import json
from datetime import datetime
from typing import Any, Mapping, MutableMapping, Optional, TypedDict, Union

import xmltodict
from botocore import xform_name
from botocore.compat import formatdate
from botocore.model import (
    ListShape,
    MapShape,
    NoShapeFoundError,
    OperationModel,
    Shape,
    StructureShape,
)
from botocore.utils import is_json_value_header, parse_to_aware_datetime

Serialized = MutableMapping[str, Any]


class ResponseDict(TypedDict):
    body: str
    headers: MutableMapping[str, str]
    status_code: int


class SerializationContext:
    def __init__(self, request_id: Optional[str] = None) -> None:
        self.request_id = request_id or "request-id"


class ErrorShape(StructureShape):
    # Overriding super class property to keep mypy happy...
    @property
    def error_code(self) -> str:
        code = str(super().error_code)
        return code


class ShapeHelpersMixin:
    @staticmethod
    def get_serialized_name(shape: Shape, default_name: str) -> str:
        return shape.serialization.get("name", default_name)

    @staticmethod
    def is_flattened(shape: Shape) -> bool:
        return shape.serialization.get("flattened", False)

    @staticmethod
    def is_http_header_trait(shape: Shape) -> bool:
        return hasattr(shape, "serialization") and shape.serialization.get(
            "location"
        ) in ["header", "headers"]

    @staticmethod
    def is_not_bound_to_body(shape: Shape) -> bool:
        return hasattr(shape, "serialization") and "location" in shape.serialization


class TimestampSerializer:
    TIMESTAMP_FORMAT_ISO8601 = "iso8601"
    TIMESTAMP_FORMAT_RFC822 = "rfc822"
    TIMESTAMP_FORMAT_UNIX = "unixtimestamp"

    ISO8601 = "%Y-%m-%dT%H:%M:%SZ"
    ISO8601_MICRO = "%Y-%m-%dT%H:%M:%S.%fZ"

    def __init__(self, default_format: str) -> None:
        self.default_format = default_format

    def serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        timestamp_format = shape.serialization.get(
            "timestampFormat", self.default_format
        )
        serialized_value = self._convert_timestamp_to_str(value, timestamp_format)
        serialized[key] = serialized_value

    def _timestamp_iso8601(self, value: datetime) -> str:
        if value.microsecond > 0:
            timestamp_format = self.ISO8601_MICRO
        else:
            timestamp_format = self.ISO8601
        return value.strftime(timestamp_format)

    @staticmethod
    def _timestamp_unixtimestamp(value: datetime) -> float:
        return int(calendar.timegm(value.timetuple()))

    def _timestamp_rfc822(self, value: Union[datetime, float]) -> str:
        if isinstance(value, datetime):
            value = self._timestamp_unixtimestamp(value)
        return formatdate(value, usegmt=True)

    def _convert_timestamp_to_str(
        self, value: Union[int, str, datetime], timestamp_format: str
    ) -> str:
        timestamp_format = timestamp_format.lower()
        converter = getattr(self, "_timestamp_%s" % timestamp_format)
        datetime_obj = parse_to_aware_datetime(value)  # type: ignore
        final_value = converter(datetime_obj)
        return final_value


class HeaderSerializer(ShapeHelpersMixin):
    # https://smithy.io/2.0/spec/http-bindings.html#httpheader-serialization-rules
    DEFAULT_ENCODING = "utf-8"
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_RFC822

    def __init__(self, **kwargs: Mapping[str, Any]) -> None:
        super().__init__(**kwargs)
        self._timestamp_serializer = TimestampSerializer(self.DEFAULT_TIMESTAMP_FORMAT)

    def serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        method = getattr(
            self, "_serialize_type_%s" % shape.type_name, self._default_serialize
        )
        method(serialized, value, shape, key)

    @staticmethod
    def _default_serialize(
        serialized: Serialized, value: Any, _: Shape, key: str
    ) -> None:
        serialized[key] = str(value)

    def _serialize_type_boolean(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        boolean_value = "true" if value else "false"
        self._default_serialize(serialized, boolean_value, shape, key)

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        list_value = ",".join(value)
        self._default_serialize(serialized, list_value, shape, key)

    def _serialize_type_map(
        self, serialized: Serialized, value: Any, shape: MapShape, _: str
    ) -> None:
        header_prefix = self.get_serialized_name(shape, "")
        for key, val in value.items():
            full_key = header_prefix + key
            self._default_serialize(serialized, val, shape, full_key)

    def _serialize_type_string(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        string_value = value
        if is_json_value_header(shape):
            json_value = json.dumps(value, separators=(",", ":"))
            string_value = self._base64(json_value)
        self._default_serialize(serialized, string_value, shape, key)

    def _serialize_type_timestamp(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        wrapper = {}
        self._timestamp_serializer.serialize(wrapper, value, shape, "timestamp")
        self._default_serialize(serialized, wrapper["timestamp"], shape, key)

    def _base64(self, value: Union[str, bytes]) -> str:
        if isinstance(value, str):
            value = value.encode(self.DEFAULT_ENCODING)
        return base64.b64encode(value).strip().decode(self.DEFAULT_ENCODING)


class ResponseSerializer(ShapeHelpersMixin):
    CONTENT_TYPE = "text"
    DEFAULT_ENCODING = "utf-8"
    DEFAULT_RESPONSE_CODE = 200
    DEFAULT_ERROR_RESPONSE_CODE = 400
    # From the spec, the default timestamp format if not specified is iso8601.
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_ISO8601
    # Clients can change this to a different MutableMapping (i.e. OrderedDict) if they want.
    # This is used in the compliance test to match the hash ordering used in the tests.
    # NOTE: This is no longer necessary because dicts post 3.6 are ordered:
    # https://stackoverflow.com/questions/39980323/are-dictionaries-ordered-in-python-3-6
    MAP_TYPE = dict

    def __init__(
        self,
        operation_model: OperationModel,
        context: Optional[SerializationContext] = None,
        pretty_print: Optional[bool] = False,
        value_picker: Any = None,
    ) -> None:
        self.operation_model = operation_model
        self.context = context or SerializationContext()
        self.pretty_print = pretty_print
        if value_picker is None:
            value_picker = self._default_value_picker
        self._value_picker = value_picker
        self._timestamp_serializer = TimestampSerializer(self.DEFAULT_TIMESTAMP_FORMAT)

    def _create_default_response(self) -> ResponseDict:
        response_dict: ResponseDict = {
            "body": "",
            "headers": {},
            "status_code": self.DEFAULT_RESPONSE_CODE,
        }
        return response_dict

    def serialize(self, result: Any) -> ResponseDict:
        resp = self._create_default_response()
        if self._is_error_result(result):
            resp = self._serialize_error(resp, result)
        else:
            resp = self._serialize_result(resp, result)
        return resp

    def _serialize_error(
        self,
        resp: ResponseDict,
        error: Exception,
    ) -> ResponseDict:
        error_shape = self._get_error_shape(error)
        serialized_error = self.MAP_TYPE()
        self._serialize_error_metadata(serialized_error, error, error_shape)
        return self._serialized_error_to_response(
            resp, error, error_shape, serialized_error
        )

    def _serialize_result(self, resp: ResponseDict, result: Any) -> ResponseDict:
        output_shape = self.operation_model.output_shape
        serialized_result = self.MAP_TYPE()
        if output_shape is not None:
            assert isinstance(output_shape, StructureShape)  # mypy hint
            self._serialize(serialized_result, result, output_shape, "")
        return self._serialized_result_to_response(
            resp, result, output_shape, serialized_result
        )

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        raise NotImplementedError("Must be implemented in subclass.")

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        raise NotImplementedError("Must be implemented in subclass.")

    def _serialize_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        raise NotImplementedError("Must be implemented in subclass.")

    def _serialize_body(self, body: Any) -> str:
        raise NotImplementedError("Must be implemented in subclass.")

    # Some extra utility methods subclasses can use.
    @staticmethod
    def _is_error_result(result: object) -> bool:
        return isinstance(result, Exception)

    def _base64(self, value: Union[str, bytes]) -> str:
        if isinstance(value, str):
            value = value.encode(self.DEFAULT_ENCODING)
        return base64.b64encode(value).strip().decode(self.DEFAULT_ENCODING)

    @staticmethod
    def _default_value_picker(obj: Any, key: str, _: Shape, default: Any = None) -> Any:
        if not hasattr(obj, "__getitem__"):
            return getattr(obj, key, default)

        try:
            return obj[key]
        except (KeyError, IndexError, TypeError, AttributeError):
            return getattr(obj, key, default)

    def _get_value(self, value: Any, key: str, shape: Shape) -> Any:
        return self._value_picker(value, key, shape)

    @staticmethod
    def _get_error_shape_name(error: Exception) -> str:
        shape_name = getattr(error, "code", error.__class__.__name__)
        return shape_name

    def _get_error_shape(self, error: Exception) -> ErrorShape:
        shape_name = self._get_error_shape_name(error)
        try:
            # TODO: there is also an errors array in the operation model,
            # but I think it only includes the possible errors for that
            # operation.  Maybe we try that first, then try all shapes?
            shape = self.operation_model.service_model.shape_for(shape_name)
            # We convert to ErrorShape to keep mypy happy...
            shape = ErrorShape(
                shape_name,
                shape._shape_model,  # type: ignore[attr-defined]
                shape._shape_resolver,  # type: ignore[attr-defined]
            )
        except NoShapeFoundError:
            generic_error_model = {
                "exception": True,
                "type": "structure",
                "members": {},
                "error": {
                    "code": shape_name,
                },
            }
            shape = ErrorShape(shape_name, generic_error_model)
        return shape

    #
    # Default serializers for the various model Shape types.
    # These can be overridden in subclasses to provide protocol-specific implementations.
    #
    def _serialize(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        method = getattr(
            self, "_serialize_type_%s" % shape.type_name, self._default_serialize
        )
        method(serialized, value, shape, key)

    @staticmethod
    def _default_serialize(
        serialized: Serialized, value: Any, _: Shape, key: str
    ) -> None:
        serialized[key] = value

    def _serialize_type_structure(
        self, serialized: Serialized, value: Any, shape: StructureShape, key: str
    ) -> None:
        if value is None:
            return
        if key:
            new_serialized = self.MAP_TYPE()
            serialized[key] = new_serialized
            serialized = new_serialized
        for member_key, member_shape in shape.members.items():
            self._serialize_structure_member(
                serialized, value, member_shape, member_key
            )

    def _serialize_structure_member(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        member_value = self._get_value(value, key, shape)
        if member_value is not None:
            key_name = self.get_serialized_name(shape, key)
            self._serialize(serialized, member_value, shape, key_name)

    def _serialize_type_map(
        self, serialized: Serialized, value: Any, shape: MapShape, key: str
    ) -> None:
        map_list = []
        if self.is_flattened(shape):
            items_name = key
            serialized[items_name] = map_list
        else:
            items_name = "entry"
            serialized[key] = {items_name: map_list}
        key_shape = shape.key
        assert isinstance(key_shape, Shape)
        value_shape = shape.value
        assert isinstance(value_shape, Shape)
        for key in value:
            wrapper = {"__current__": {}}
            key_prefix = self.get_serialized_name(key_shape, "key")
            value_prefix = self.get_serialized_name(value_shape, "value")
            self._serialize(wrapper["__current__"], key, key_shape, key_prefix)
            self._serialize(
                wrapper["__current__"], value[key], value_shape, value_prefix
            )
            map_list.append(wrapper["__current__"])

    def _serialize_type_timestamp(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        value_wrapper = {}
        value_key = "timestamp"
        self._timestamp_serializer.serialize(value_wrapper, value, shape, value_key)
        self._default_serialize(serialized, value_wrapper[value_key], shape, key)

    def _serialize_type_blob(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        blob_value = self._base64(value)
        self._default_serialize(serialized, blob_value, shape, key)


class BaseJSONSerializer(ResponseSerializer):
    APPLICATION_AMZ_JSON = "application/x-amz-json-{version}"
    DEFAULT_TIMESTAMP_FORMAT = "unixtimestamp"

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        resp["body"] = self._serialize_body(serialized_result)
        resp["headers"]["Content-Type"] = self._get_protocol_specific_content_type()
        return resp

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        resp["body"] = self._serialize_body(serialized_error)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        error_code = self._get_protocol_specific_error_code(shape.error_code)
        resp["headers"]["X-Amzn-Errortype"] = error_code
        resp["headers"]["Content-Type"] = self._get_protocol_specific_content_type()
        return resp

    def _get_protocol_specific_content_type(self) -> str:
        content_type = self.CONTENT_TYPE
        service_model = self.operation_model.service_model
        protocol = service_model.protocol
        if protocol == "json":
            json_version = service_model.metadata.get("jsonVersion", "1.0")
            content_type = self.APPLICATION_AMZ_JSON.format(version=json_version)
        return content_type

    def _get_protocol_specific_error_code(
        self,
        error_code: str,
    ) -> str:
        # https://smithy.io/2.0/aws/protocols/aws-json-1_1-protocol.html#operation-error-serialization
        service_metadata = self.operation_model.service_model.metadata
        json_version = service_metadata.get("jsonVersion")
        prefix = service_metadata.get("targetPrefix")
        if json_version == "1.0" and prefix is not None:
            error_code = prefix + "#" + error_code
        return error_code

    def _serialize_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        error_code = self._get_protocol_specific_error_code(shape.error_code)
        serialized["__type"] = error_code
        message = getattr(error, "message", None) or str(error)
        if shape is not None:
            self._serialize(serialized, error, shape, "")
        if message:
            serialized["Message"] = message

    def _serialize_body(self, body: Mapping[str, Any]) -> str:
        body_encoded = json.dumps(body, indent=4 if self.pretty_print else None)
        return body_encoded

    def _serialize_type_map(
        self, serialized: Serialized, value: Any, shape: MapShape, key: str
    ) -> None:
        map_obj = self.MAP_TYPE()
        serialized[key] = map_obj
        for sub_key, sub_value in value.items():
            assert isinstance(shape.value, Shape)  # mypy hint
            self._serialize(map_obj, sub_value, shape.value, sub_key)

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        list_obj = []
        serialized[key] = list_obj
        for list_item in value:
            wrapper = {}
            # The JSON list serialization is the only case where we aren't
            # setting a key on a dict.  We handle this by using
            # a __current__ key on a wrapper dict to serialize each
            # list item before appending it to the serialized list.
            assert isinstance(shape.member, Shape)  # mypy hint
            self._serialize(wrapper, list_item, shape.member, "__current__")
            if "__current__" in wrapper:
                list_obj.append(wrapper["__current__"])
            else:
                list_obj.append(list_item)

    def _serialize_type_structure(
        self, serialized: Serialized, value: Any, shape: StructureShape, key: str
    ) -> None:
        if shape.is_document_type:
            serialized[key] = value
            return
        super()._serialize_type_structure(serialized, value, shape, key)


class BaseXMLSerializer(ResponseSerializer):
    CONTENT_TYPE = "text/xml"

    def _serialize_namespace_attribute(self, serialized: Serialized) -> None:
        if (
            self.CONTENT_TYPE == "text/xml"
            and "xmlNamespace" in self.operation_model.metadata
        ):
            namespace = self.operation_model.metadata["xmlNamespace"]
            serialized["@xmlns"] = namespace

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        error_wrapper = {
            "ErrorResponse": {
                "Error": serialized_error,
                "RequestId": self.context.request_id,
            }
        }
        self._serialize_namespace_attribute(error_wrapper["ErrorResponse"])
        resp["body"] = self._serialize_body(error_wrapper)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        result_key = f"{self.operation_model.name}Result"
        result_wrapper = {
            result_key: serialized_result,
        }
        self._serialize_namespace_attribute(result_wrapper[result_key])
        resp["body"] = self._serialize_body(result_wrapper)
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialize_error_metadata(
        self,
        serialized: Serialized,
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        sender_fault = shape.metadata.get("error", {}).get("senderFault", True)
        serialized["Type"] = "Sender" if sender_fault else "Receiver"
        serialized["Code"] = shape.error_code
        message = getattr(error, "message", None)
        if message is not None:
            serialized["Message"] = message
        # Serialize any error model attributes.
        self._serialize(serialized, error, shape, "")

    def _serialize_body(self, body: Serialized) -> str:
        body_encoded = xmltodict.unparse(
            body,
            full_document=False,
            short_empty_elements=True,
            pretty=self.pretty_print,
        )
        return body_encoded

    #
    # https://smithy.io/2.0/aws/protocols/aws-query-protocol.html#xml-shape-serialization
    #
    def _serialize_type_boolean(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        # We're slightly more permissive here than we should be because the
        # moto backends are not consistent in how they store boolean values.
        # TODO: This should eventually be turned into a strict `is True` check.
        boolean_conditions = [
            (value is True),
            (str(value).lower() == "true"),
        ]
        boolean_value = "true" if any(boolean_conditions) else "false"
        self._default_serialize(serialized, boolean_value, shape, key)

    def _serialize_type_integer(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        integer_value = int(value)
        self._default_serialize(serialized, integer_value, shape, key)

    def _serialize_type_list(
        self, serialized: Serialized, value: Any, shape: ListShape, key: str
    ) -> None:
        assert isinstance(shape.member, Shape)  # mypy hinting
        list_obj = []
        if self.is_flattened(shape):
            items_name = self.get_serialized_name(shape.member, key)
            serialized[items_name] = list_obj
        else:
            items_name = self.get_serialized_name(shape.member, "member")
            serialized[key] = {items_name: list_obj}
        for list_item in value:
            wrapper = {}
            self._serialize(wrapper, list_item, shape.member, "__current__")
            list_obj.append(wrapper["__current__"])
        if not list_obj:
            serialized[key] = ""

    _serialize_type_long = _serialize_type_integer

    def _serialize_type_string(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        string_value = str(value)
        self._default_serialize(serialized, string_value, shape, key)


class BaseRestSerializer(ResponseSerializer):
    EMPTY_BODY: Serialized = ResponseSerializer.MAP_TYPE()
    REQUIRES_EMPTY_BODY = False

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        if "payload" in serialized_result:
            # Payload trumps all and is delivered as-is.
            resp["body"] = serialized_result["payload"]
        else:
            if not serialized_result["body"]:
                if self.REQUIRES_EMPTY_BODY:
                    resp["body"] = self._serialize_body(self.EMPTY_BODY)
            else:
                resp = super()._serialized_result_to_response(
                    resp, result, shape, serialized_result.get("body", {})
                )
        if "headers" in serialized_result:
            resp["headers"].update(serialized_result["headers"])
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialize_result(self, resp: ResponseDict, result: Any) -> ResponseDict:
        output_shape = self.operation_model.output_shape
        serialized_result = {
            "body": {},
            "headers": {},
        }
        if output_shape is not None:
            assert isinstance(output_shape, StructureShape)
            self._serialize(serialized_result, result, output_shape, "")
            payload_member = output_shape.serialization.get("payload")
            if payload_member is not None:
                payload_shape = output_shape.members[payload_member]
                payload_value = self._get_value(result, payload_member, payload_shape)
                self._serialize_payload(serialized_result, payload_value, payload_shape)

        return self._serialized_result_to_response(
            resp, result, output_shape, serialized_result
        )

    def _serialize_payload(
        self,
        serialized: Serialized,
        payload: Any,
        payload_shape: Shape,
    ) -> None:
        if payload_shape.type_name in ["blob", "string"]:
            # If it's streaming, then the body is just the value of the payload.
            serialized["payload"] = payload
        else:
            # If there's a payload member, we serialize only that one member to the body.
            serialized["body"] = self.MAP_TYPE()
            self._serialize(serialized["body"], payload, payload_shape, "")

    def _serialize_structure_member(
        self, serialized: Serialized, value: Any, shape: Shape, key: str
    ) -> None:
        if self.is_not_bound_to_body(shape):
            if self.is_http_header_trait(shape) and "headers" in serialized:
                member_value = self._get_value(value, key, shape)
                if member_value is not None:
                    key_name = self.get_serialized_name(shape, key)
                    header_serializer = HeaderSerializer()
                    header_serializer.serialize(
                        serialized["headers"], member_value, shape, key_name
                    )
        elif "body" in serialized:
            if not serialized["body"]:
                serialized["body"] = self.MAP_TYPE()
            # we're at the top-level structure
            super()._serialize_structure_member(serialized["body"], value, shape, key)
        else:
            # we're in nested structure
            super()._serialize_structure_member(serialized, value, shape, key)


class RestXMLSerializer(BaseRestSerializer, BaseXMLSerializer):
    DEFAULT_TIMESTAMP_FORMAT = TimestampSerializer.TIMESTAMP_FORMAT_ISO8601


class RestJSONSerializer(BaseRestSerializer, BaseJSONSerializer):
    CONTENT_TYPE = "application/json"
    REQUIRES_EMPTY_BODY = True


class JSONSerializer(BaseJSONSerializer):
    pass


class QuerySerializer(BaseXMLSerializer):
    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        error_wrapper = {
            "ErrorResponse": {
                "Error": serialized_error,
                "RequestId": self.context.request_id,
            }
        }
        self._serialize_namespace_attribute(error_wrapper["ErrorResponse"])
        resp["body"] = self._serialize_body(error_wrapper)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: Optional[StructureShape],
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        response_key = f"{self.operation_model.name}Response"
        response_wrapper = {response_key: {}}
        if shape is not None:
            result_key = shape.serialization.get("resultWrapper", f"{shape.name}Result")
            response_wrapper[response_key][result_key] = serialized_result
        response_wrapper[response_key]["ResponseMetadata"] = {
            "RequestId": self.context.request_id
        }
        self._serialize_namespace_attribute(response_wrapper[response_key])
        resp["body"] = self._serialize_body(response_wrapper)
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp


class EC2Serializer(QuerySerializer):
    def _serialize_body(self, body: Mapping[str, Any]) -> str:
        body_serialized = xmltodict.unparse(
            body,
            full_document=True,
            pretty=self.pretty_print,
            short_empty_elements=True,
        )
        return body_serialized

    def _serialize_error_metadata(
        self,
        serialized: MutableMapping[str, Any],
        error: Exception,
        shape: ErrorShape,
    ) -> None:
        serialized["Code"] = shape.error_code
        message = getattr(error, "message", None)
        if message is not None:
            serialized["Message"] = message
        # Serialize any error model attributes.
        self._serialize(serialized, error, shape, "")

    def _serialized_error_to_response(
        self,
        resp: ResponseDict,
        error: Exception,
        shape: ErrorShape,
        serialized_error: MutableMapping[str, Any],
    ) -> ResponseDict:
        error_wrapper = {
            "Response": {
                "Errors": [{"Error": serialized_error}],
                "RequestID": self.context.request_id,
            }
        }
        self._serialize_namespace_attribute(error_wrapper["Response"])
        resp["body"] = self._serialize_body(error_wrapper)
        status_code = shape.metadata.get("error", {}).get(
            "httpStatusCode", self.DEFAULT_ERROR_RESPONSE_CODE
        )
        resp["status_code"] = status_code
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp

    def _serialized_result_to_response(
        self,
        resp: ResponseDict,
        result: Any,
        shape: StructureShape,
        serialized_result: MutableMapping[str, Any],
    ) -> ResponseDict:
        response_key = f"{self.operation_model.name}Response"
        result_wrapper = {
            response_key: serialized_result,
        }
        result_wrapper[response_key]["requestId"] = self.context.request_id
        self._serialize_namespace_attribute(result_wrapper[response_key])
        resp["body"] = self._serialize_body(result_wrapper)
        resp["headers"]["Content-Type"] = self.CONTENT_TYPE
        return resp


SERIALIZERS = {
    "ec2": EC2Serializer,
    "json": JSONSerializer,
    "query": QuerySerializer,
    "rest-json": RestJSONSerializer,
    "rest-xml": RestXMLSerializer,
}


class XFormedAttributePicker:
    """Can be injected into a ResponseSerializer to aid in plucking AWS model
    attributes specified in `camelCase` or `PascalCase` from Python objects
    with standard `snake_case` attribute names.

    For a model attribute named `DBInstanceIdentifier`, this class will check
    for the following attributes on the provided object:
       * `DBInstanceIdentifier`
       * `db_instance_identifier`
    If the provided object is a class named `DBInstance`, this class will also
    check for the following attribute on the provided object:
       * `identifier`

    Uses ``botocore.xform_name`` to translate the attribute name.
    """

    def __init__(self) -> None:
        self._xform_cache = {}

    def __call__(self, value: Any, key: str, shape: Shape) -> Any:
        return self._get_value(value, key, shape)

    def xform_name(self, name: str) -> str:
        return xform_name(name, _xform_cache=self._xform_cache)

    def _get_value(self, value: Any, key: str, shape: Shape) -> Any:
        new_value = None
        possible_keys = [key, self.xform_name(key)]
        # If a class `Role` has an attribute named `arn`, that will work for a `RoleArn` key.
        if hasattr(value, "__class__"):
            class_name = value.__class__.__name__
            if key.lower().startswith(class_name.lower()):
                short_key = key[len(class_name) :]
                if short_key:  # Will be empty string if class name same as key
                    possible_keys.append(self.xform_name(short_key))
        if shape is not None:
            serialization_key = shape.serialization.get("name", key)
            if serialization_key != key:
                possible_keys.append(serialization_key)
        for key in possible_keys:
            new_value = ResponseSerializer._default_value_picker(value, key, shape)
            if new_value is not None:
                break
        return new_value



import json

import boto3
import requests

from moto import mock_aws, settings
from moto.s3.utils import bucket_and_name_from_url

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_invoke_endpoint__default_results():
    client = boto3.client("sagemaker-runtime", region_name="ap-southeast-1")
    body = client.invoke_endpoint(
        EndpointName="asdf", Body="qwer", Accept="sth", TargetModel="tm"
    )

    assert body["Body"].read() == b"body"
    assert body["CustomAttributes"] == "custom_attributes"


@mock_aws
def test_invoke_endpoint():
    client = boto3.client("sagemaker-runtime", region_name="us-east-1")
    base_url = (
        "localhost:5000" if settings.TEST_SERVER_MODE else "motoapi.amazonaws.com"
    )

    sagemaker_result = {
        "results": [
            {
                "Body": "first body",
                "ContentType": "text/xml",
                "InvokedProductionVariant": "prod",
                "CustomAttributes": "my_attr",
            },
            {"Body": "second body"},
        ]
    }
    requests.post(
        f"http://{base_url}/moto-api/static/sagemaker/endpoint-results",
        json=sagemaker_result,
    )

    # Return the first item from the list
    body = client.invoke_endpoint(EndpointName="asdf", Body="qwer")
    assert body["Body"].read() == b"first body"

    # Same input -> same output
    body = client.invoke_endpoint(EndpointName="asdf", Body="qwer")
    assert body["Body"].read() == b"first body"

    # Different input -> second item
    body = client.invoke_endpoint(
        EndpointName="asdf", Body="qwer", Accept="sth", TargetModel="tm"
    )
    assert body["Body"].read() == b"second body"


@mock_aws
def test_invoke_endpoint_async():
    client = boto3.client("sagemaker-runtime", region_name="us-east-1")
    base_url = (
        "localhost:5000" if settings.TEST_SERVER_MODE else "motoapi.amazonaws.com"
    )

    sagemaker_result = {
        "results": [
            {"data": json.dumps({"first": "output"})},
            {
                "is_failure": True,
                "data": "second failure",
            },
        ]
    }
    requests.post(
        f"http://{base_url}/moto-api/static/sagemaker/async-endpoint-results",
        json=sagemaker_result,
    )

    # Return the first item from the list
    body = client.invoke_endpoint_async(EndpointName="asdf", InputLocation="qwer")
    first_output_location = body["OutputLocation"]
    first_failure_location = body["FailureLocation"]

    # Same input -> same output
    body = client.invoke_endpoint_async(EndpointName="asdf", InputLocation="qwer")
    assert body["OutputLocation"] == first_output_location
    assert body["FailureLocation"] == first_failure_location

    s3 = boto3.client("s3", "us-east-1")
    bucket_name, obj = bucket_and_name_from_url(first_output_location)
    resp = s3.get_object(Bucket=bucket_name, Key=obj)
    resp = json.loads(resp["Body"].read().decode("utf-8"))
    assert resp == {"first": "output"}

    # Different input -> second item
    body = client.invoke_endpoint_async(
        EndpointName="asdf", InputLocation="asf", InferenceId="sth"
    )
    second_failure_location = body["FailureLocation"]
    assert body["InferenceId"] == "sth"

    bucket_name, obj = bucket_and_name_from_url(second_failure_location)
    resp = s3.get_object(Bucket=bucket_name, Key=obj)
    resp = resp["Body"].read().decode("utf-8")
    assert resp == "second failure"


@mock_aws
def test_invoke_endpoint_async_should_read_sync_queue_if_async_not_configured():
    client = boto3.client("sagemaker-runtime", region_name="us-east-1")
    base_url = (
        "localhost:5000" if settings.TEST_SERVER_MODE else "motoapi.amazonaws.com"
    )

    sagemaker_result = {
        "results": [
            {"Body": "support sync queue for backward compatibility"},
        ]
    }
    requests.post(
        f"http://{base_url}/moto-api/static/sagemaker/endpoint-results",
        json=sagemaker_result,
    )

    # Return the first item from the list
    body = client.invoke_endpoint_async(EndpointName="asdf", InputLocation="qwer")
    first_output_location = body["OutputLocation"]
    first_failure_location = body["FailureLocation"]

    # Same input -> same output
    body = client.invoke_endpoint_async(EndpointName="asdf", InputLocation="qwer")
    assert body["OutputLocation"] == first_output_location
    assert body["FailureLocation"] == first_failure_location

    s3 = boto3.client("s3", "us-east-1")
    bucket_name, obj = bucket_and_name_from_url(first_output_location)
    resp = s3.get_object(Bucket=bucket_name, Key=obj)
    resp = json.loads(resp["Body"].read().decode("utf-8"))
    assert resp["Body"] == "support sync queue for backward compatibility"



from moto.dynamodb.utils import find_duplicates, find_path_overlaps


def test_find_duplicates_simple_duplicate():
    assert find_duplicates(["a", "a"]) == ["a", "a"]


def test_find_duplicates_no_duplicates():
    assert find_duplicates(["a", "b"]) == []


def test_find_duplicates_out_of_order():
    assert find_duplicates(["b", "a", "b"]) == ["b", "b"]


def test_find_path_overlaps_simple_overlap():
    assert find_path_overlaps(["a", "a.b"]) == ["a", "a.b"]


def test_find_path_overlaps_no_overlap():
    assert find_path_overlaps(["a", "b"]) == []


def test_find_path_overlaps_reverse_overlap():
    assert find_path_overlaps(["a.b", "a"]) == ["a.b", "a"]



import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_aws
def test_describe_account_settings():
    client = boto3.client("quicksight", region_name="us-east-2")
    resp = client.describe_account_settings(AwsAccountId=ACCOUNT_ID)
    assert resp["AccountSettings"]["AccountName"] == "default"
    assert resp["AccountSettings"]["Edition"] == "STANDARD"
    assert resp["AccountSettings"]["PublicSharingEnabled"] is False
    assert resp["AccountSettings"]["TerminationProtectionEnabled"] is False


@mock_aws
def test_update_account_settings():
    client = boto3.client("quicksight", region_name="eu-west-1")
    client.update_account_settings(
        AwsAccountId=ACCOUNT_ID,
        DefaultNamespace="default",
        NotificationEmail="test@moto.com",
        TerminationProtectionEnabled=True,
    )

    resp = client.describe_account_settings(AwsAccountId=ACCOUNT_ID)
    assert resp["AccountSettings"]["TerminationProtectionEnabled"] is True
    assert resp["AccountSettings"]["NotificationEmail"] == "test@moto.com"

    client.update_account_settings(
        AwsAccountId=ACCOUNT_ID,
        DefaultNamespace="default",
        NotificationEmail="test2@moto.com",
    )
    resp = client.describe_account_settings(AwsAccountId=ACCOUNT_ID)
    assert resp["AccountSettings"]["TerminationProtectionEnabled"] is True
    assert resp["AccountSettings"]["NotificationEmail"] == "test2@moto.com"


@mock_aws
def test_update_public_sharing_settings():
    client = boto3.client("quicksight", region_name="us-east-2")
    client.update_public_sharing_settings(
        AwsAccountId=ACCOUNT_ID, PublicSharingEnabled=True
    )
    resp = client.describe_account_settings(AwsAccountId=ACCOUNT_ID)
    assert resp["AccountSettings"]["PublicSharingEnabled"] is True

    client.update_public_sharing_settings(
        AwsAccountId=ACCOUNT_ID, PublicSharingEnabled=False
    )
    resp = client.describe_account_settings(AwsAccountId=ACCOUNT_ID)
    assert resp["AccountSettings"]["PublicSharingEnabled"] is False
